import ssl
import socket
from typing import Tuple
from hyper.common.util import to_native_string

from urllib.parse import urlparse
from hyper import HTTP11Connection, HTTPConnection
from hyper.common.bufsocket import BufferedSocket
from hyper.common.exceptions import TLSUpgrade
from hyper.contrib import HTTP20Adapter
from hyper.tls import init_context

from tcp_tls_tunnel.utils import generate_basic_header, generate_proxy_url
from tcp_tls_tunnel.dto import ProxyOptions, AdapterOptions, TunnelOptions
from tcp_tls_tunnel.exceptions import ProxyError


def _create_tunnel(tunnel_opts: TunnelOptions,
                   dest_host: str,
                   dest_port: int,
                   server_name: str = None,
                   proxy: ProxyOptions = None,
                   timeout: int = None) -> Tuple[socket.socket, str]:
    """
    Sends CONNECT method to a proxy and returns a socket with established
    connection to the target.
    :returns: socket, proto
    """

    headers = {
        "Authorization": generate_basic_header(tunnel_opts.auth_login, tunnel_opts.auth_password),
        "Client": tunnel_opts.client.value,
        "Connection": 'keep-alive',
        "Server-Name": server_name or dest_host,
        "Host": tunnel_opts.host,
        "Secure": str(int(tunnel_opts.secure)),
        "HTTP2": str(int(tunnel_opts.http2)),
    }

    if proxy:
        headers["Proxy"] = generate_proxy_url(proxy=proxy)

    conn = HTTP11Connection(tunnel_opts.host, tunnel_opts.port, timeout=timeout)
    conn.request('CONNECT', '%s:%d' % (dest_host, dest_port),
                 headers=headers)

    resp = conn.get_response()

    try:
        proto = resp.headers.get("Alpn-Protocol")[0].decode('utf-8')
    except TypeError:
        proto = 'http/1.1'

    if resp.status != 200:
        raise ProxyError(
            "Tunnel connection failed: %d %s" %
            (resp.status, to_native_string(resp.reason)),
            response=resp
        )

    return getattr(conn, "_sock"), proto


class TunnelHTTP20Adapter(HTTP20Adapter):
    def __init__(self,
                 adapter_opts: AdapterOptions,
                 proxy_opts: ProxyOptions = None,
                 window_manager=None,
                 *args, **kwargs):
        super(TunnelHTTP20Adapter, self).__init__(window_manager=window_manager, *args, **kwargs)
        self.adapter_opts = adapter_opts
        self.proxy_opts = proxy_opts

    def get_connection(self, host, port, scheme, cert=None, verify=True,
                       proxy=None, timeout=None):

        """
        Gets an appropriate HTTP/2 connection object based on
        host/port/scheme/cert tuples.
        """
        secure = (scheme == 'https')

        if port is None:  # pragma: no cover
            port = 80 if not secure else 443

        ssl_context = None
        if not verify:
            verify = False
            ssl_context = init_context(cert=cert)
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE
        elif verify is True and cert is not None:
            ssl_context = init_context(cert=cert)
        elif verify is not True:
            ssl_context = init_context(cert_path=verify, cert=cert)

        if proxy:
            proxy_headers = self.proxy_headers(proxy)
            proxy_netloc = urlparse(proxy).netloc
        else:
            proxy_headers = None
            proxy_netloc = None

        # We put proxy headers in the connection_key, because
        # ``proxy_headers`` method might be overridden, so we can't
        # rely on proxy headers being the same for the same proxies.
        proxy_headers_key = (frozenset(proxy_headers.items())
                             if proxy_headers else None)
        connection_key = (host, port, scheme, cert, verify,
                          proxy_netloc, proxy_headers_key)
        try:
            conn = self.connections[connection_key]
        except KeyError:
            conn = TunnelHTTPConnection(
                self.adapter_opts,
                self.proxy_opts,
                host=host,
                port=port,
                secure=secure,
                ssl_context=ssl_context,
                proxy_host=proxy_netloc,
                proxy_headers=proxy_headers,
                timeout=timeout,
            )
            self.connections[connection_key] = conn
        return conn


class TunnelHTTPConnection(HTTPConnection):
    def __init__(self,
                 adapter_opts: AdapterOptions,
                 proxy_opts: ProxyOptions = None,
                 host=None,
                 port=None,
                 secure=None,
                 enable_push=False,
                 ssl_context=None,
                 proxy_host=None,
                 proxy_port=None,
                 proxy_headers=None,
                 timeout=None,
                 **kwargs):
        super().__init__(host=host,
                         port=port,
                         secure=secure,
                         enable_push=enable_push,
                         ssl_context=ssl_context,
                         proxy_host=proxy_host,
                         proxy_port=proxy_port,
                         proxy_headers=proxy_headers,
                         timeout=timeout,
                         **kwargs)

        self._conn = TunnelHTTP11Connection(
            tunnel_opts=TunnelOptions(
                host=adapter_opts.host,
                port=adapter_opts.port,
                auth_login=adapter_opts.auth_login,
                auth_password=adapter_opts.auth_password,
                client=adapter_opts.client,
                secure=secure,
                http2=True
            ),
            proxy_opts=proxy_opts,
            host=self._host,
            port=self._port,
            **self._h1_kwargs
        )

    def __exit__(self, type, value, tb):  # pragma: no cover
        self._conn.close()
        return False


class TunnelHTTP11Connection(HTTP11Connection):
    def __init__(self,
                 tunnel_opts: TunnelOptions,
                 proxy_opts: ProxyOptions = None,
                 host=None, port=None, secure=None, ssl_context=None,
                 proxy_host=None, proxy_port=None, proxy_headers=None,
                 timeout=None,
                 **kwargs):
        super(TunnelHTTP11Connection, self).__init__(host=host, port=port,
                                                     secure=secure, ssl_context=ssl_context,
                                                     proxy_host=proxy_host, proxy_port=proxy_port,
                                                     proxy_headers=proxy_headers, timeout=timeout,
                                                     **kwargs)
        self.tunnel_opts = tunnel_opts
        self.proxy_opts = proxy_opts
        self.timeout = timeout

    def connect(self):
        """
        Connect to the server specified when the object was created. This is a
        no-op if we're already connected.
        :returns: Nothing.
        """

        if self._sock is None:

            # Tunnel socket creation with tunnel's TLS proto
            sock, proto = _create_tunnel(
                tunnel_opts=self.tunnel_opts,
                proxy=self.proxy_opts,
                dest_host=self.host,
                dest_port=self.port,
                server_name=None,  # TODO: server_name
                timeout=self.timeout
            )

            sock = BufferedSocket(sock, self.network_buffer_size)
            sock.settimeout(self.timeout)  # Set read timeout

            if self.secure is not True:
                proto = 'http/1.1'

            if proto not in ('http/1.1', None):
                raise TLSUpgrade(proto, sock)

            self._sock = sock
        return



# Create our own ConnectionError.
try:  # pragma: no cover
    ConnectionError = ConnectionError
except NameError:  # pragma: no cover
    class ConnectionError(Exception):
        """
        An error occurred during connection to a host.
        """


class ProxyError(ConnectionError):
    """
    An error occurred during connection to a proxy.
    """
    def __init__(self, message, response):
        self.response = response
        super(ProxyError, self).__init__(message)


from base64 import b64encode

from tcp_tls_tunnel.dto import ProxyOptions


def generate_basic_header(login: str, password: str):
    auth_string = f"{login}:{password}".encode("utf-8")
    auth_token = b64encode(auth_string).decode("ascii")
    return f"Basic {auth_token}"


def generate_proxy_url(proxy: ProxyOptions):
    if proxy:
        proxy_host_port: str = f"{proxy.host}:{proxy.port}"

        if proxy.auth_login and proxy.auth_password:
            return (
                f"{proxy.auth_login}:"
                f"{proxy.auth_password}@"
                f"{proxy_host_port}"
            )
        else:
            return proxy_host_port
    return None



import enum


class Client(enum.Enum):
    CHROME = "CHROME"
    CHROME_58 = "CHROME_58"
    CHROME_62 = "CHROME_62"
    CHROME_70 = "CHROME_70"
    CHROME_72 = "CHROME_72"
    CHROME_83 = "CHROME_83"

    FIREFOX = "FIREFOX"
    FIREFOX_55 = "FIREFOX_55"
    FIREFOX_56 = "FIREFOX_56"
    FIREFOX_63 = "FIREFOX_63"
    FIREFOX_65 = "FIREFOX_65"

    IOS = "IOS"
    IOS_11_1 = "IOS_11_1"
    IOS_12_1 = "IOS_12_1"



import logging
import socket
from ssl import SSLContext
from typing import Union, Tuple

from httpcore import SyncConnectionPool
from httpcore._backends.sync import SyncBackend, SyncSocketStream
from httpcore._sync.connection import SyncHTTPConnection
from httpx import HTTPTransport, create_ssl_context, Limits
from httpx._config import DEFAULT_LIMITS
from httpx._types import VerifyTypes, CertTypes

from tcp_tls_tunnel.dto import AdapterOptions, ProxyOptions, TunnelOptions
from tcp_tls_tunnel.requests_adapter import create_tunnel_connection

logging.basicConfig(level=logging.WARNING)


class TunnelHTTPTransport(HTTPTransport):
    def __init__(
            self,
            adapter_opts: AdapterOptions,
            proxy_opts: ProxyOptions = None,
            verify: VerifyTypes = True,
            cert: CertTypes = None,
            http1: bool = True,
            limits: Limits = DEFAULT_LIMITS,
            trust_env: bool = True,
            retries: int = 0,
            backend: str = "sync",
    ) -> None:
        super(TunnelHTTPTransport, self).__init__()
        ssl_context = create_ssl_context(verify=verify, cert=cert, trust_env=trust_env)

        self._pool = TunnelHTTP20SyncConnectionPool(
            adapter_opts=adapter_opts,
            proxy_opts=proxy_opts,
            ssl_context=ssl_context,
            http1=http1,
            http2=True,
            retries=retries,
            max_connections=limits.max_connections,
            max_keepalive_connections=limits.max_keepalive_connections,
            keepalive_expiry=limits.keepalive_expiry,
            backend=backend
        )


class TunnelSyncSocketStream(SyncSocketStream):

    def __init__(self, sock: socket.socket, proto: str):
        super(TunnelSyncSocketStream, self).__init__(sock=sock)
        self.proto = proto

    def get_http_version(self) -> str:
        return "HTTP/2" if self.proto == "h2" else "HTTP/1.1"


class TunnelHTTP20SyncConnectionPool(SyncConnectionPool):

    def __init__(
            self,
            adapter_opts: AdapterOptions,
            proxy_opts: ProxyOptions,
            ssl_context: SSLContext = None,
            max_connections: int = None,
            max_keepalive_connections: int = None,
            keepalive_expiry: float = None,
            http1: bool = True,
            http2: bool = False,
            uds: str = None,
            local_address: str = None,
            retries: int = 0,
            max_keepalive: int = None,
            backend: Union[SyncBackend, str] = "sync",
    ):
        super(TunnelHTTP20SyncConnectionPool, self).__init__(
            ssl_context=ssl_context,
            max_connections=max_connections,
            max_keepalive_connections=max_keepalive_connections,
            keepalive_expiry=keepalive_expiry,
            http1=http1,
            http2=http2,
            uds=uds,
            local_address=local_address,
            retries=retries,
            max_keepalive=max_keepalive,
            backend=backend
        )
        self.adapter_opts = adapter_opts
        self.proxy_opts = proxy_opts

    def _create_connection(
        self,
        origin: Tuple[bytes, bytes, int],
    ) -> SyncHTTPConnection:
        host, port = origin[1].decode("utf-8"), origin[2]
        secure = int("https" == origin[0].decode("utf-8"))

        conn, proto = create_tunnel_connection(
            tunnel_opts=TunnelOptions(
                host=self.adapter_opts.host,
                port=self.adapter_opts.port,
                auth_login=self.adapter_opts.auth_login,
                auth_password=self.adapter_opts.auth_password,
                client=self.adapter_opts.client,
                secure=True if secure == 1 else False,
                http2=True
            ),
            dest_host=host,
            dest_port=port,
            proxy=self.proxy_opts,
            server_name=None  # TODO: server_name
        )

        return SyncHTTPConnection(
            origin=origin,
            http1=self._http1,
            http2=self._http2,
            keepalive_expiry=self._keepalive_expiry,
            uds=self._uds,
            ssl_context=self._ssl_context,
            local_address=self._local_address,
            socket=TunnelSyncSocketStream(sock=conn.sock, proto=proto),
            retries=self._retries,
            backend=self._backend,
        )



from dataclasses import dataclass
from tcp_tls_tunnel.constants import Client


@dataclass
class AdapterOptions:
    host: str
    port: int

    # auth options
    auth_login: str = None
    auth_password: str = None

    # headers options
    client: Client = Client.CHROME


@dataclass
class TunnelOptions(AdapterOptions):
    secure: bool = False
    http2: bool = False


@dataclass
class ProxyOptions:
    host: str
    port: int

    # auth_options
    auth_login: str = None
    auth_password: str = None



from typing import Tuple
import http
import socket
from http import client
from tcp_tls_tunnel.utils import generate_basic_header, generate_proxy_url

from tcp_tls_tunnel.dto import TunnelOptions, ProxyOptions


class TunnelHTTPConnection(client.HTTPConnection):
    def __init__(self,
                 host,
                 port=None,
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                 source_address=None,
                 blocksize=8192):
        super(TunnelHTTPConnection, self).__init__(
            host=host,
            port=port,
            timeout=timeout,
            source_address=source_address,
            blocksize=blocksize
        )
        self.tunnel_response_headers = {}

    def _tunnel(self):
        self.tunnel_response_headers.clear()
        connect_str = "CONNECT %s:%d HTTP/1.0\r\n" % (self._tunnel_host, self._tunnel_port)
        connect_bytes = connect_str.encode("ascii")
        self.send(connect_bytes)
        for header, value in self._tunnel_headers.items():
            header_str = "%s: %s\r\n" % (header, value)
            header_bytes = header_str.encode("latin-1")
            self.send(header_bytes)
        self.send(b'\r\n')

        response = self.response_class(self.sock, method=self._method)
        (version, code, message) = response._read_status()

        if code != http.HTTPStatus.OK:
            self.close()
            raise OSError("Tunnel connection failed: %d %s" % (code, message.strip()))

        self.tunnel_response_headers = client.parse_headers(response.fp)


def create_tunnel_connection(tunnel_opts: TunnelOptions,
                             dest_host: str,
                             dest_port: int,
                             server_name: str = None,
                             proxy: ProxyOptions = None) -> Tuple[TunnelHTTPConnection, str]:

    conn = TunnelHTTPConnection(tunnel_opts.host, tunnel_opts.port)
    headers = {
        "Authorization": generate_basic_header(tunnel_opts.auth_login, tunnel_opts.auth_password),
        "Client": tunnel_opts.client.value,
        "Connection": 'keep-alive',
        "Server-Name": server_name or dest_host,
        "Host": tunnel_opts.host,
        "Secure": str(int(tunnel_opts.secure)),
        "HTTP2": str(int(tunnel_opts.http2)),
    }

    if proxy:
        headers["Proxy"] = generate_proxy_url(proxy=proxy)

    conn.set_tunnel(dest_host, port=dest_port, headers=headers)
    conn.connect()

    try:
        proto = conn.tunnel_response_headers.get("Alpn-Protocol")
    except TypeError:
        proto = 'http/1.1'

    return conn, proto



from urllib.parse import urlparse

import requests

import requests.cookies
import requests.utils
from requests.structures import CaseInsensitiveDict
from requests.adapters import BaseAdapter

from tcp_tls_tunnel.dto import AdapterOptions, TunnelOptions, ProxyOptions
from tcp_tls_tunnel.create_tunnel_connection import create_tunnel_connection


class TunneledHTTPAdapter(BaseAdapter):

    def __init__(self,
                 adapter_opts: AdapterOptions,
                 proxy_opts: ProxyOptions = None):
        super(BaseAdapter, self).__init__()
        self.adapter_opts = adapter_opts
        self.proxy = proxy_opts

    def close(self):
        pass

    def send(self, request, **kwargs):
        parsed_url = urlparse(request.url)
        secure = False

        if parsed_url.port:
            destination_port = parsed_url.port
        elif parsed_url.scheme == 'https':
            secure = True
            destination_port = 443
        elif parsed_url.scheme == 'http':
            destination_port = 80
        else:
            raise ValueError("Unexpected url protocol. You can specify port in the url.")

        connection, proto = create_tunnel_connection(
            tunnel_opts=TunnelOptions(
                host=self.adapter_opts.host,
                port=self.adapter_opts.port,
                auth_login=self.adapter_opts.auth_login,
                auth_password=self.adapter_opts.auth_password,
                client=self.adapter_opts.client,
                secure=secure,
                http2=False
            ),
            dest_host=parsed_url.hostname,
            dest_port=destination_port,
            server_name=None,  # TODO: server_name
            proxy=self.proxy
        )

        try:
            connection.request(method=request.method,
                               url=request.url,
                               body=request.body,
                               headers=request.headers)
            r = connection.getresponse()
            # Response preparation process
            response = requests.Response()
            response.status_code = r.status
            response.headers = CaseInsensitiveDict(r.headers)
            response.raw = r
            response.reason = r.reason
            response.url = request.url
            response.request = request
            response.connection = connection

            # Added content to response
            # (under the hood it will set r.text and allow r.json() operation)
            response._content = r.read()
            response.encoding = requests.utils.get_encoding_from_headers(
                response.headers
            )
            requests.cookies.extract_cookies_to_jar(
                response.cookies,
                request,
                r
            )
        finally:
            connection.close()

        return response



import pathlib
from setuptools import setup, find_packages

BASE_DIR = pathlib.Path(__file__).parent
README = (BASE_DIR / "README.md").read_text()


setup(
    name="tcp-tls-tunnel",
    version='1.0.0',
    description="TCP TLS tunnel for HTTP requests with HTTP2 support.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Virtual-Shard/tcp-tls-tunnel-py",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    include_package_data=True,
    packages=find_packages(exclude=("tests",)),
    extras_require={
        "httpx": [
            "h2>=4.0,<5.0",
            "httpx>=0.18.2"
        ],
        "hyper": [
            "h2>=2.6.2,<3.0",
            "hpack>=3.0.0,<4.0",
            "hyper>=0.7.0,<0.8.0"
        ],
    }
)
